from typing import Any, List

try:
    # This works on MacOS but not Ubuntu
    from markdowntable import markdownTable  # type: ignore
except:
    # This works on Ubuntu but not MacOS
    from markdownTable import markdownTable  # type: ignore

from pydantic import BaseModel

from py2tf.converters.jinja import JinjaConverter


class MarkdownTableBuilder(markdownTable):
    def getHeader(self):
        # The only change in this function from the parent is the `title` casing
        header = ""
        if self.row_sep in ("topbottom", "always"):
            header += self.newline_char + self.var_row_sep_last + self.newline_char
        for key in self.data[0].keys():
            margin = self.var_padding[key] - len(key)
            right = self.getMargin(margin)
            header += "|" + key.title().rjust(self.var_padding[key] - right, self.padding_char).ljust(
                self.var_padding[key], self.padding_char
            )
        header += "|" + self.newline_char
        if self.row_sep == "always":
            header += self.var_row_sep + self.newline_char
        if self.row_sep == "markdown":
            header += self.var_row_sep.replace("+", "|") + self.newline_char
        return header


class MarkdownConverter(JinjaConverter):

    numbers: List[str] = []  # ["int", "float", "integer"]
    objects: List[str] = ["dict", "object"]
    pass_through_fields: List[str] = ["string", "int", "float", "integer"]

    def __init__(self, format="table") -> None:
        self.format = format

    def convert(self, model: BaseModel) -> str:

        schema = model.schema()
        output = ""
        fields = []
        for field_name, field_data in schema["properties"].items():
            if field_name == "type":
                continue
            fields.append(self.convert_field(schema, field_name, field_data))

        # We need at least one field to make a table.
        if len(fields) < 1:
            return ""

        # Sort required fields to the top, then sort by alphabet.
        fields.sort(key=lambda t: (0 if t["Required"] == "Yes" else 1, t["Name"]))

        return MarkdownTableBuilder(fields).setParams(quote=False, row_sep="markdown").getMarkdown()

    def convert_field(self, schema, field_name, field_data) -> Any:
        field_data = self.get_field_data(schema, field_name, field_data)
        description = ""
        if "description" in field_data and field_data["description"]:
            description = field_data["description"]

        return {
            "Name": f"`{field_name}`",
            "Type": field_data.get("type", ""),
            "Description": description,
            "Default": field_data["default"] if field_data["has_default"] else "",
            "Required": "Yes" if field_data["required"] else "No",
        }

    def get_type(self, field_name, field_data):
        type, validators = super(MarkdownConverter, self).get_type(field_name, field_data)
        if type == "bool":
            type = "boolean"
        if "list" in type:
            type = type.replace("list", "array")
        return type, validators



from importlib import import_module
from typing import Union

import typer

from .converters.jinja import JinjaConverter
from .converters.markdown import MarkdownConverter

app = typer.Typer()


@app.command()
def main(model_path, format: str = typer.Option("hcl")):
    module_path, class_name = model_path.split(":", 2)
    module = import_module(module_path)
    model = getattr(module, class_name)

    converter: Union[JinjaConverter, MarkdownConverter]
    if format == "markdown":
        converter = MarkdownConverter(format=format)
    else:
        converter = JinjaConverter(format=format)

    print(converter.convert(model))


if __name__ == "__main__":
    app()



import json
import os
import sys
import textwrap
from pathlib import Path
from typing import Any, Dict, List, cast

from glom import glom  # type: ignore
from jinja2 import Environment, FileSystemLoader, Template
from pydantic import BaseModel

from py2tf import validators

env = Environment(loader=FileSystemLoader(Path(__file__).parent.parent / "templates"))

SETTINGS = {
    "hcl": {
        "template": cast(Template, env.get_template("hcl.tf")),
        "delimiter": "\n\n",
    },
    "tfvars": {
        "template": cast(Template, env.get_template("variable.tfvars")),
        "delimiter": "\n",
    },
    "module": {
        "template": cast(Template, env.get_template("module_params.tf")),
        "delimiter": "\n",
        "wrapper": cast(Template, env.get_template("module_wrapper.tf")),
    },
}


class JinjaConverter:

    numbers: List[str] = ["int", "float", "integer"]
    objects: List[str] = ["dict", "object"]
    pass_through_fields: List[str] = ["string"]

    def __init__(self, format="hcl") -> None:
        if format not in SETTINGS:
            raise ValueError(f"Format must be hcl or tfvars")
        self.template = SETTINGS[format]["template"]
        self.delimiter = SETTINGS[format]["delimiter"]
        self.wrapper = cast(Template, SETTINGS[format]["wrapper"]) if "wrapper" in SETTINGS[format] else None

    def convert(self, model: BaseModel) -> str:
        schema = model.schema()
        output = ""
        for field_name, field_data in schema["properties"].items():
            output += self.convert_field(schema, field_name, field_data)
            output += self.delimiter  # type: ignore

        if self.wrapper:
            output = self.wrapper.render(parameters=output)

        return output.strip()

    def _get_reference(self, schema, reference):
        # Convert reference into glom format.
        path = reference.strip("#/").replace("/", ".")
        return glom(schema, path)

    def convert_field(self, schema, field_name, field_data) -> Any:
        field_data = self.get_field_data(schema, field_name, field_data)
        return self.template.render(**field_data)  # type: ignore

    def get_field_data(self, schema, field_name, field_data) -> Dict[str, Any]:
        # This field references another field. Swap the field_data out.
        if "$ref" in field_data:
            field_data = self._get_reference(schema, field_data["$ref"])

        if "allOf" in field_data:
            for parent_model in field_data["allOf"]:
                if "$ref" in parent_model:
                    parent_model = self._get_reference(schema, parent_model["$ref"])
                field_data |= parent_model

        type, type_rules = self.get_type(field_name, field_data)

        # Since "null" and "false" are valid defaults we can't rely on default being defined.
        # To get around that we have `has_default`.

        has_default = False
        default = None
        required = field_name in schema.get("required", {})
        if "default" in field_data or not required:
            has_default = True
            default = self.get_default(type, field_data.get("default"))

        rules = self.get_validation_list(field_name, field_data, allow_none=not required)
        if type_rules:
            rules += type_rules

        return {
            "name": field_name,
            "type": type,
            "description": field_data.get("description"),
            "has_default": has_default,
            "default": default,
            "sensitive": "writeOnly" in field_data,
            "required": required,
            "validation_rules": rules,
        }

    def get_type(self, field_name, field_data):

        type_validators = []

        if "enum" in field_data:
            if "type" not in field_data:
                field_data["type"] = "string"
            type_validators.append(validators.oneof(field_name, field_data["enum"], allow_none=False))

        if "type" not in field_data:
            print(field_data)
            raise ValueError(f"Unknown type for {field_name}")

        field_type = field_data["type"].lower()

        if field_type in self.pass_through_fields:
            return field_type, type_validators

        if field_type in self.numbers:
            return "number", type_validators

        if field_type == "array":
            subtype = self.get_subtype(field_name, field_data)
            return f"list({subtype})", None

        if field_type in self.objects:
            subtype = self.get_subtype(field_name, field_data)
            return f"map({subtype})", None

        if field_type == "boolean":
            return "bool", None

        return "string", None

    def get_subtype(self, field_name, field_data):
        subtype = "any"
        if "items" in field_data:
            if "type" in field_data["items"]:
                subtype, rules = self.get_type(field_name, field_data["items"])
        if "additionalProperties" in field_data and "type" in field_data["additionalProperties"]:
            subtype = field_data["additionalProperties"]["type"]
        return subtype

    def get_default(self, type, value):
        if type.startswith("list"):
            return textwrap.indent(json.dumps(value, indent=2), "  ").strip() if value else "[]"
        if type.startswith("map"):
            return textwrap.indent(json.dumps(value, indent=2), "  ").strip() if value else "{}"

        # Custom Nulls above this line

        if value == None:
            return "null"
        if type == "number":
            return value
        if type == "string":
            escaped_value = str(value).replace('"', '\\"')
            return f'"{escaped_value}"'
        if type == "bool":
            return "true" if value else "false"

        return "null"

    def get_validation_list(self, field_name, field_data, allow_none):
        tf_rules = []
        for key, value in field_data.items():
            if not hasattr(validators, key.lower()):
                continue
            rule = getattr(validators, key.lower())(field_name, value, allow_none)
            if rule:
                tf_rules.append(rule)

        return tf_rules



import json
import re


class Rule:
    condition: str
    message: str
    comment: str

    def __init__(self, condition, message, rule_name, var_name, allow_none) -> None:
        if allow_none:
            self.condition = f"var.{var_name} == null ? true : {condition}"
        else:
            self.condition = condition
        self.message = message
        self.comment = f"Automatically Generated from Rule `{rule_name}` with allow_none set to {str(allow_none)}"


"""
The function names in this file all correspond with the attributes in
JSONSchema, which in turn are returned by pydantic. Each function returns a
Terraform conditional and an error message for when that condition is not
matched.
"""


def exclusiveminimum(name, value, allow_none):
    return Rule(
        condition=f"var.{name} > {value}",
        message=f"Field should be larger than {value}",
        rule_name="gt",
        var_name=name,
        allow_none=allow_none,
    )


def minimum(name, value, allow_none):
    return Rule(
        condition=f"var.{name} >= {value}",
        message=f"Field should be greater than or equal to {value}",
        rule_name="ge",
        var_name=name,
        allow_none=allow_none,
    )


def exclusivemaximum(name, value, allow_none):
    return Rule(
        condition=f"var.{name} < {value}",
        message=f"Field should be smaller than {value}",
        rule_name="lt",
        var_name=name,
        allow_none=allow_none,
    )


def maximum(name, value, allow_none):
    return Rule(
        condition=f"var.{name} <= {value}",
        message=f"Field should be less than or equal to {value}",
        rule_name="le",
        var_name=name,
        allow_none=allow_none,
    )


def multipleof(name, value, allow_none):
    return Rule(
        condition=f"var.{name} % {value} == 0",
        message=f"Field should be a multiple of {value}",
        rule_name="multipleof",
        var_name=name,
        allow_none=allow_none,
    )


def minitems(name, value, allow_none):
    return Rule(
        condition=f"length(var.{name}) >= {value}",
        message=f"Field should contain at least {value} items",
        rule_name="minitems",
        var_name=name,
        allow_none=allow_none,
    )


def maxitems(name, value, allow_none):
    return Rule(
        condition=f"length(var.{name}) <= {value}",
        message=f"Field should contain at most {value} items",
        rule_name="maxitems",
        var_name=name,
        allow_none=allow_none,
    )


def minlength(name, value, allow_none):
    return Rule(
        condition=f"length(var.{name}) >= {value}",
        message=f"Field should not be less than {value} characters",
        rule_name="minlength",
        var_name=name,
        allow_none=allow_none,
    )


def maxlength(name, value, allow_none):
    return Rule(
        condition=f"length(var.{name}) <= {value}",
        message=f"Field should not be larger than {value} characters",
        rule_name="maxlength",
        var_name=name,
        allow_none=allow_none,
    )


def uniqueitems(name, value, allow_none):
    return Rule(
        condition=f"length(distinct(var.{name})) == length(var.{name})",
        message=f"Field should only contain unique items",
        rule_name="uniqueitems",
        var_name=name,
        allow_none=allow_none,
    )


def pattern(name, value, allow_none):
    # Terraform needs the pattern as a string, which means we need to escape slashes.
    pattern = make_regex_simple(value).replace("\\", "\\\\")
    return Rule(
        condition=f'length(regexall("{pattern}", var.{name})) > 0',
        message=f"Field does not match regex pattern {pattern}",
        rule_name="pattern",
        var_name=name,
        allow_none=allow_none,
    )


def oneof(name, value, allow_none):
    pretty_list = ", ".join([str(x) for x in value])
    terraform_list = json.dumps(value)
    return Rule(
        condition=f"contains({terraform_list}, var.{name})",
        message=f"Field should be one of: {pretty_list}",
        rule_name="oneof",
        var_name=name,
        allow_none=allow_none,
    )


# https://stackoverflow.com/a/14919203/212774
# https://creativecommons.org/licenses/by-sa/4.0/
def make_regex_simple(verbose):
    WS_RX = r"(?<!\\)((\\{2})*)\s+"
    CM_RX = r"(?<!\\)((\\{2})*)#.*$(?m)"

    return re.sub(WS_RX, "\\1", re.sub(CM_RX, "\\1", verbose))
